#!/usr/bin/env python
''' File to perform simple scan-to-map alignment and map creation.'''
from __future__ import division
import roslib; roslib.load_manifest('mrol_mapping')
import rospy

import mrol_mapping.poseutil as poseutil
import mrol_mapping.mrol as mrol
import mrol_mapping.util as util
import mrol_mapping.occupiedlist as occupiedlist
import mrol_mapping.mapper as mapper
import mrol_mapping.pointcloud as pointcloud
import numpy as np
import rosutils_mrol.rosutils_mrol as rosutils
from sensor_msgs.msg import PointCloud2
from thread import allocate_lock
import tf
import sys

class RosMapper:
    def __init__(self,num_base_scans=2,use_ICP=False):
        rospy.loginfo("Initializing")
        # can use mrol's inherent alignment method, or ICP
        self.use_ICP = use_ICP
        if self.use_ICP:
            rospy.loginfo("Using ICP for alignment")
            import icp_mrol.icp_mrol as icp
            # make an ICP instance to do the alignment
            self.icp = icp.ICP(reciprocal=False, visualise=False)
        # resolution determines how fine the map is and also how 
        # accurate the alignment is, but finer = slower
        self.res = 0.04#0.05
        # levels determines region of convergence, more = larger region 
        # of convergences. smaller is faster, but may not converge
        self.levels = 4
        # minimal overlap for point cloud to be added to map
        self.min_overlap =0.50
        # variable initialisation
        self.bestpose = poseutil.Pose3D()
        # mapper init
        #self.mapper = mapper.VoxelMap(resolution=self.res, levels=self.levels)
        self.mapper = mapper.VoxelMap(resolution=self.res, levels=self.levels, useraddfn=colour_add)
        # keep track of scan number
        self.counter = 0
        # lock to have thread-safetiness
        self.lock = allocate_lock()
        self.num_base_scans = num_base_scans
        # publish whole map to /mrol
        self.pub = rospy.Publisher("/mrol", PointCloud2)
        #listener = tf.TransformListener()
        self.br = tf.TransformBroadcaster()

    def PCcallback(self, msg):
        ######################
        # Point cloud received
        ######################
        rospy.logdebug("Recieved point cloud")
        # converting to xyz array
        xyzs = rosutils.pc22xyzs(msg, max_range=3.)
        # converting to MROL PointCloud format
        xyzs = pointcloud.PointCloud(xyzs)    
        
        # We can sub-sample here to make it run faster for the Kinect. 
        # But this is not as accurate as using full data set.
        #xyzs.volumetricsample(self.res)
        # Could also do feature extraction here instead
        
        # re-align x fwd, z up
        xyzs.points[:,:3] = poseutil.transformPointsTuple(xyzs.points[:,:3], [0,0,0,-np.pi/2.,0,-np.pi/2.]) 
    
        # get transformations between world and robot at time of message
        # to have a guess for the position
        #try:
        #    # rospy.Time(0) means last one
        #    (trans, rot) = listener.lookupTransform(\
        #            'world', 'robot', rospy.Time(0))
        #    rot = tf.transformations.euler_from_quaternion(rot)
        #    pose = np.r_[trans, rot]
        #    self.bestpose = poseutil.Pose3D(pose)
        #except (tf.LookupException, tf.ConnectivityException):
        #    rospy.logerr("Couldn't get transformation update - taking last one")
        
        #################################
        # run MROL mapper to achieve SLAM
        #################################
        # NOTE: ROS subscribtions run in parallel! The user needs to care about
        # thread-safetiness - but the MROL methods are not implemented
        # thread-safe, so a lock is needed here
        self.lock.acquire()

        #####################
        # localization in map
        #####################
        # first scan is used straight away - all other scans are aligned 
        # to fit best to the map
        if self.counter is not 0:
            rospy.logdebug("Aligning new points to map")
            if self.use_ICP:
                if len(xyzs) > len(self.mapper.get_points()):
                    xyzs.points = util.getfastsample(xyzs.points, len(self.mapper.get_points())-1)
                #import pdb;pdb.set_trace()    
                new_pose = self.icp.align(self.bestpose.getTuple(), np.float64(xyzs.points[:,:3]), self.mapper.get_points().points[:,:3], terminate_threshold=0.01, reject_threshold=0.5)[0]
                #new_pose = self.icp.align(self.bestpose.getTuple(), np.float64(xyzs.points[:,:3]), self.mapper.get_points().points[:,:3], terminate_threshold=0.01, reject_threshold=0.5, point2plane=True)[0] # need better normal estimates for this to work
                self.bestpose = poseutil.Pose3D(new_pose)
            else:
                self.bestpose = self.mapper.align_points(xyzs, self.bestpose)[0]

        ###################################
        # Mapping if points fit good enough
        ###################################
        # get overlap of best alignment
        ol = self.mapper.get_overlap(self.bestpose.getTuple(), xyzs.points[:,:3])
        rospy.logdebug("overlap is %5.1f%%, minimum is %5.1f%%",\
                ol/len(xyzs) * 100, self.min_overlap * 100)

        # adding the first self.num_base_scans of scans without check, the rest only
        # if the overlap is big enough
        if self.counter < self.num_base_scans or ol/len(xyzs) > self.min_overlap:
            rospy.logdebug("Adding points to map")
            xyzs.volumetricsample(self.res) # adding is slow, so downsampling speeds it up a fair bit!
            self.mapper.add_points(xyzs, self.bestpose)
        
        #TODO: use quaternion_from_euler(self.bestpose.getTuple()[3:] instead?
        quat = tf.transformations.quaternion_from_matrix(self.bestpose.getMatrix())
        self.counter += 1
        self.lock.release()

        #TODO: use other transformation here
        self.br.sendTransform(self.bestpose.getTuple()[:3],
                quat,
                rospy.Time.now(),
                "/openni_camera",
                "/world")
        rospy.logdebug("TF sent")
        # publish point cloud
        #points = self.mapper.get_points().points
        #pointsPC = self.mapper.get_points()
        #pointsPC.display()
        #points = pointsPC.points
        voxel_data, voxel_res = self.mapper.get_voxels(level=-1)
        pointsROS = np.hstack([voxel_data[:,:3]*voxel_res, voxel_data[:,4:]])
        pcROS = rosutils.xyzs2pc2(pointsROS, msg.header.stamp, self.counter, frame_id='/world')
        self.pub.publish(pcROS)
        rospy.logdebug("Finished processing point cloud")
        
        
def colour_add(cur_data,data):
    ''' An example of a user add function for incrementing user data in 
    a map voxel given a new measurement'''
    # First we must handle the case where this is the first time we've 
    # added data to the voxel, easily identified because the current
    # voxel data only contains the number of prior observations (zero).
    if len(cur_data) == 0:
        return data
    else:
        # This should be customised to the user's application. Typically 
        # it should check for consistency in data length also
        return np.int64(np.array(data)) # some merging algorithm
        
def run(ICP):
    ##########################
    # Initialize ROS interface
    ##########################
    rospy.init_node('Kinect_mapper_demo',log_level=rospy.DEBUG )
    # sleep a little - this avoids some ROS bugs
    rospy.sleep(2)
    orm = RosMapper(num_base_scans=2,use_ICP=ICP)
    # define how many scan should be without minimum overlay
    # this has to be at least 1, otherwise no map can be created
    rospy.Subscriber("/camera/rgb/points", PointCloud2, orm.PCcallback, queue_size=5) 
    rospy.spin()
    
def profile(): # Doesn't work!
    import cProfile
    cProfile.runctx('run()',globals(),locals(),'out.profile')

if __name__ == '__main__':
    ICP = '-ICP' in sys.argv
    try:
        run(ICP)
    except rospy.ROSInterruptException: pass
   
