''' Author: Julian Ryde and Nick Hillier'''
# Establish correspondences
# Estimate transformation parameters using mean square cost function
# alternative just use an optimisation toolbox
# Transform points with improved pose
# reiterate
import roslib
roslib.load_manifest('mrol_mapping')
import rospy

import numpy as np

import mrol_mapping.poseutil as poseutil
import mrol_mapping.occupiedlist as occupiedlist


# TODO work out why "import pyflann" doesn't work
import sys
import subprocess
def get_shell_response(cmd):
    p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
    stdout, stderr = p.communicate()
    return stdout


def import_pyflann():
    try:
        print sys.path[-1]
        import pyflann
        return pyflann.FLANN()
    except (OSError, ImportError):
        print "No flann install found, trying custom paths..."
    flann_paths = [get_shell_response('rospack find flann')[:-1]+'/python', 
                   get_shell_response('rospack find mrol_mapping')[:-1]+'/3rdParty/flann-1.2-linux-i686/python',
                   get_shell_response('rospack find mrol_mapping')[:-1]+'/3rdParty/flann-1.2-linux-x86_64/python',
                   get_shell_response('rospack find mrol_mapping')[:-1]+'/3rdParty/flann-1.6.11-src/src/python']
    for flann_path in flann_paths:
        sys.path.append(flann_path)
        try:
            print sys.path[-1]
            import pyflann
            break
        except (OSError, ImportError):
            sys.path.pop()

    return pyflann.FLANN()

def eigen_normal(pt):
    if pt.shape[0] < 3:
        return np.array([0,0,0])
    cov = np.cov(pt.T)
    eigs = np.linalg.eig(cov)
    sortind = np.argsort(eigs[0])
    # check planarity
    if eigs[0][sortind[0]]/eigs[0][sortind[1]] < 0.25:
        return eigs[1][:,sortind[0]]
    else:
        return np.array([0,0,0])

class ICP():
    def __init__(self, reciprocal=False, visualise=False):
        self.reciprocal = reciprocal
        self.flann = import_pyflann()
        if self.reciprocal:
            self.flann2 = import_pyflann()
        self.visualise = visualise
        self.Qbuilt = False
        self.patch_method = 'voxels' # 'voxels' or 'nearest'
        self.vox_res = 5. # size of voxels to use for computing surface normals if 'voxel' method
        self.patch_number = 20 # must be > 3 # number of neighbouring points to use for computing surface normals if 'nearest' method
        
    def __calcnormal(self, Q, Qinds):
        QindsN, QdistsN = self.flann.nn_index(Q[Qinds], num_neighbors=3)
        QindsN = QindsN[:,1:]
        QdistsN = QdistsN[:,1:]
        vec1 = Q[Qinds] - Q[QindsN[:,0]]
        vec2 = Q[Qinds] - Q[QindsN[:,1]]
        Qnormals = np.cross(vec1,vec2)
        # TODO: push this loop to cython or do it some other way?
        for i, norm in enumerate(Qnormals):
            Qnormals[i,:] = norm/(np.sqrt(np.sum(norm**2)))
            if np.isnan(Qnormals[i,:]).any():
                #import pdb;pdb.set_trace()
                # gets in here when vec1 or vec2 = [0,0,0]
                Qnormals[i,:] = np.array([0,0,0])
        return Qnormals
        
    def __calcnormalpatch(self, Q, Qinds):
        QindsN, QdistsN = self.flann.nn_index(Q[Qinds], num_neighbors=self.patch_number)
        QindsN = QindsN[:,1:]
        QdistsN = QdistsN[:,1:]
        pts = Q[QindsN]
        Qnormals = []
        for i, pt in enumerate(pts):
            Qnormals.append(eigen_normal(pt))
        return np.asarray(Qnormals)

    def __calcnormalvoxelise(self, Q, Qinds):
        points = Q[Qinds]
        centers = occupiedlist.pointstovoxels(points, self.vox_res)
        keys = occupiedlist._pack(centers)
        voxels = {}
        for i, key in enumerate(keys):
            if not voxels.has_key(key):
                voxels[key] = []
            voxels[key].append(points[i])
        #sorted_inds = np.argsort(keys)
        #keys_sorted = keys[sorted_inds]
        #centers_sorted = centers[sorted_inds]
        #diffs = np.diff(keys_sorted)
        Qnormals = []
        for i, key in enumerate(keys):
            normal = eigen_normal(np.asarray(voxels[key]))
            # force normals to point towards z
            #if normal[2] < 0:
            #    normal = -normal 
            Qnormals.append(normal)
        return np.asarray(Qnormals)
        
        #import pydb;pydb.set_trace()
        #return eigen_normals(voxels)

    def __associate(self, P, Q, threshold, calc_normal=False):
        '''
        Q is typically a large point cloud (the truth) and P is 
        typically a single scan.
        Establishes those points in Q that best associate with or 
        correspond to the points in P.  
        Best is currently judged based on proximity, thresholded by 
        the threshold parameter.
        Need to call build index before this. 
        Points further than threshold are considered to have no corresponding 
        points.
        '''
        assert len(Q) >= len(P), 'The static point cloud must have more or equal points to that against which it is being associated'
        
        Qinds, Pdists = self.flann.nn_index(P)
        Qnormals = None
        
        if not self.reciprocal:
            # Pinds selects those points in P with a partner in Q
            # Julian believes dists is the square of the distance but is not sure
            # TODO flann library supports rejection thresholds Julian wonders
            Dinds = Pdists < threshold**2
            Qinds = Qinds[Dinds]
            Pinds = np.cumsum(np.ones(len(P),dtype=np.int64)) - 1 # 0 based indexing!
            Pinds = Pinds[Dinds]
            if calc_normal:
                if self.patch_method == 'nearest':
                    Qnormals = self.__calcnormalpatch(Q, Qinds)
                elif self.patch_method == 'voxels':
                    Qnormals = self.__calcnormalvoxelise(Q, Qinds)
                else:
                    Qnormals = self.__calcnormal(Q, Qinds)

            Qpts, Ppts = self.get_assoc_pts(P, Q, Pinds, Qinds)
            dists = self.get_assoc_dists(Ppts, Qpts, Qnormals)
            #dists = Pdists[Dinds]
        else:
            Pinds, Qdists = self.flann2.nn_index(Q)
            
            Ppairs = np.zeros((Pinds.shape[0],2),dtype=np.int64)

            # TODO fix these loops to something faster!
            for i, ind in enumerate(Pinds): 
                Ppairs[i,0] = i
                Ppairs[i,1] = ind
                
            Qpairs = np.zeros((Qinds.shape[0],2),dtype=np.int64)
            for i, ind in enumerate(Qinds): 
                Qpairs[i,1] = i
                Qpairs[i,0] = ind
             
            # get only those associations that are common both ways 
            # (reciprocal nearest neighbours)
            
            Qpairs_set = set((d[0], d[1]) for d in Qpairs )
            Ppairs_set = set((d[0], d[1]) for d in Ppairs )
            pairs_set = Ppairs_set & Qpairs_set # intersection
            pairs = np.array([list(e) for e in pairs_set])

            Qinds = pairs[:,0]
            Pinds = pairs[:,1]
            #dists = Qdists[Qinds]
            Qpts, Ppts = self.get_assoc_pts(P, Q, Pinds, Qinds)
            dists = self.get_assoc_dists(Ppts, Qpts, Qnormals)
            good_assocs = dists < threshold**2
            
            if calc_normal:
                if self.patch_method == 'nearest':
                    Qnormals = self.__calcnormalpatch(Q, Qinds)
                elif self.patch_method == 'voxels':
                    Qnormals = self.__calcnormalvoxelise(Q, Qinds)
                else:
                    Qnormals = self.__calcnormal(Q, Qinds)
                Qnormals = Qnormals[good_assocs]
            
            Pinds = Pinds[good_assocs]
            Qinds = Qinds[good_assocs]
            Qpts, Ppts = self.get_assoc_pts(P, Q, Pinds, Qinds)
            dists = self.get_assoc_dists(Ppts, Qpts, Qnormals)
            #dists = Pdists[good_assocs]
            
        return Qinds, Pinds, dists, Qnormals # Qnormals = None if not calc_normal
        
    def associate(self, P, Q, threshold, calc_normal=False):
        self.Qbuilt = False # force rebuild in case data's changed
        self.__buildindex(Q, P)
        return self.__associate(P, Q, threshold, calc_normal=calc_normal)
        
    def get_assoc_pts(self, P, Q, Pinds, Qinds):
        Ppts = P[Pinds]
        Qpts = Q[Qinds]
        return Qpts, Ppts
        
    def get_assoc_dists(self, Ppts, Qpts, Qnormals=None):
        if Qnormals != None :
            dists = []
            for i, normal in enumerate(Qnormals):
                dists.append(np.sqrt(np.dot((Ppts[i] - Qpts[i]),normal)**2)) # distance along normal between associated points
            dists = np.asarray(dists)
        else:
            dists = np.sqrt(np.sum((Ppts - Qpts) ** 2, axis=1)) # euclidean distance between associated points
        return dists
            
    def __buildindex(self, Q, P=None):
        if not self.Qbuilt: # only need to build the index for Q once, as it is static
            self.params = self.flann.build_index(Q)
            self.Qbuilt = True
        if self.reciprocal:
            assert P!=None
            self.params2 = self.flann2.build_index(P)
    
    def __initvisualiser(self):
        import mrol_mapping.visualiser.robotvisualiser as robotvisualiser
        self.visualise = True
        self.vis = robotvisualiser.Visualiser(title='ICP Visualiser')
        self.vis.clear()
        self.vis.setautominmax()
        
    def __vis_clear_lines(self, line_handles):
        for line in line_handles:
            # possible memory leak! How to dlete rather than just make invisible?
            line.pos[0] = [0,0,0] 
            line.pos[1] = [0,0,0] 
    
    def __calctransform(self, P, Q):
        ''' Assumes the points in P and Q are associated '''
        assert P.shape == Q.shape
        # A, B matrices symbolically produced from icp_derviation.py
        # multiply translation by n because derviation is for one point (should be n**2?)
        px, py, pz = np.sum(P, axis=0)
        px2, py2, pz2 = np.sum(P**2, axis=0)
        pxpy = np.sum(P[:, 0] * P[:, 1])
        pxpz = np.sum(P[:, 0] * P[:, 2])
        pypz = np.sum(P[:, 1] * P[:, 2])
        n = P.shape[0]
        A = np.array((
        [py2 + pz2,     -pxpy,     -pxpz,   0, -pz,  py],
        [    -pxpy, px2 + pz2,     -pypz,  pz,   0, -px],
        [    -pxpz,     -pypz, px2 + py2, -py,  px,   0],
        [        0,        pz,       -py,   n,   0,   0],
        [      -pz,         0,        px,   0,   n,   0],
        [       py,       -px,         0,   0,   0,   n]
        ))

        qx, qy, qz = np.sum(Q, axis=0)

        pzqy = np.sum(P[:, 2] * Q[:, 1])
        pyqz = np.sum(P[:, 1] * Q[:, 2])
        pxqz = np.sum(P[:, 0] * Q[:, 2])
        pzqx = np.sum(P[:, 2] * Q[:, 0])
        pyqx = np.sum(P[:, 1] * Q[:, 0])
        pxqy = np.sum(P[:, 0] * Q[:, 1])
        
        B = np.array((
        [pzqy - pyqz],
        [pxqz - pzqx],
        [pyqx - pxqy],
        [    px - qx],
        [    py - qy],
        [    pz - qz]
        ))

        X = np.linalg.solve(A, -B)
        # X is rotx, roty, rotz, x, y, z
        return np.squeeze(np.vstack((X[3:], X[:3])))
        
    def __calctransform_normal(self, P, Q, N):
        ''' Assumes the points in P and Q are associated and N is a 
        vector of normals to the Q surface at the associated points'''
        assert P.shape == Q.shape
        assert P.shape == N.shape
        # A, B matrices symbolically produced from icp_derviation_p2plane.py
        # multiply translation by n because derviation is for one point (should be n**2?)
        px, py, pz = np.sum(P, axis=0)
        px2, py2, pz2 = np.sum(P**2, axis=0)
        nx, ny, nz = np.sum(N, axis=0)
        nx2, ny2, nz2 = np.sum(N**2, axis=0)
        n = P.shape[0]
        pxpy = np.sum(P[:, 0] * P[:, 1])
        pxpz = np.sum(P[:, 0] * P[:, 2])
        pypz = np.sum(P[:, 1] * P[:, 2])
        nxny = np.sum(N[:, 0] * N[:, 1])
        nxnz = np.sum(N[:, 0] * N[:, 2])
        nynz = np.sum(N[:, 1] * N[:, 2])
        
        A = np.array((
        [           -2*nynz*pypz + ny2*pz2 + nz2*py2, nxnz*pypz + nynz*pxpz - nxny*pz2 - pxpy*nz2, nxny*pypz + nynz*pxpy - nxnz*py2 - pxpz*ny2,   nxnz*py - nxny*pz,   nynz*py - pz*ny2,  -nynz*pz + py*nz2],
        [nxnz*pypz + nynz*pxpz - nxny*pz2 - pxpy*nz2,            -2*nxnz*pxpz + nx2*pz2 + nz2*px2, nxny*pxpz + nxnz*pxpy - nynz*px2 - pypz*nx2,   -nxnz*px + pz*nx2,  nxny*pz - nynz*px,   nxnz*pz - px*nz2],
        [nxny*pypz + nynz*pxpy - nxnz*py2 - pxpz*ny2, nxny*pxpz + nxnz*pxpy - nynz*px2 - pypz*nx2,            -2*nxny*pxpy + nx2*py2 + ny2*px2,    nxny*px - py*nx2,  -nxny*py + px*ny2,  nynz*px - nxnz*py],
        [                          nxnz*py - nxny*pz,                           -nxnz*px + pz*nx2,                            nxny*px - py*nx2,               n*nx2,             n*nxny,             n*nxnz],
        [                           nynz*py - pz*ny2,                           nxny*pz - nynz*px,                            nxny*py + px*ny2,              n*nxny,              n*ny2,             n*nynz],
        [                          -nynz*pz + py*nz2,                            nxnz*pz - px*nz2,                            nynz*px - nxnz*py,             n*nx*nz,             n*nynz,              n*nz2]
        ))

        qx, qy, qz = np.sum(Q, axis=0)
        
        pzqx = np.sum(P[:, 2] * Q[:, 0])
        pyqx = np.sum(P[:, 1] * Q[:, 0])
        pxqx = np.sum(P[:, 0] * Q[:, 0])
        pzqy = np.sum(P[:, 2] * Q[:, 1])
        pyqy = np.sum(P[:, 1] * Q[:, 1])
        pxqy = np.sum(P[:, 0] * Q[:, 1])
        pzqz = np.sum(P[:, 2] * Q[:, 2])
        pyqz = np.sum(P[:, 1] * Q[:, 2])
        pxqz = np.sum(P[:, 0] * Q[:, 2])
        
        B = np.array((
        [nxny*pzqx + nxnz*pxpy + nynz*pzqz - nxny*pxpz - nxnz*pyqx - nynz*pyqy + nynz*py2 + pypz*nz2 + pzqy*ny2 - nynz*pz2 - pypz*ny2 - pyqz*nz2],
        [nxny*pypz + nxnz*pxqx + nynz*pxqy - nxny*pzqy - nxnz*pzqz - nynz*pxpy + nxnz*pz2 + pxpz*nx2 + pxqz*nz2 - nxnz*px2 - pxpz*nz2 - pzqx*nx2],
        [nxny*pyqy + nxnz*pyqz + nynz*pxpz - nxny*pxqx - nxnz*pypz - nynz*pxqz + nxny*px2 + pxpy*ny2 + pyqx*nx2 - nxny*py2 - pxpy*nx2 - pxqy*ny2],
        [                                                                                nxny*py + nxnz*pz - nxny*qy - nxnz*qz + px*nx2 - qx*nx2],
        [                                                                                nxny*px + nynz*pz - nxny*qx - nynz*qz + py*ny2 - qy*ny2],
        [                                                                                nxnz*px + nynz*py - nxnz*qx - nynz*qy + pz*nz2 - qz*nz2]
        ))

        X = np.linalg.solve(A, -B)
        # X is rotx, roty, rotz, x, y, z
        return np.squeeze(np.vstack((X[3:], X[:3])))

    def align(self, initialpose, P, Q, terminate_threshold=0.05, reject_threshold=0.5, iterations=50, point2plane=False, quiet=True, reassoc=True):
        '''Given two sets of corresponding points (P, Q) and an initial guess 
        transform calculate the transform from P to Q. (i.e. P is scan, Q is map)'''
        assert(type(P[0][0]) == type(Q[0][0])), 'P and Q must be of the same data type'
        P = poseutil.transformPoints(P, poseutil.mat(initialpose))
        # looping without re-associating does improve speed 
        # but I think it is negligible, compared to error due to poor
        # associations
        initialposemat = poseutil.mat(initialpose)
        bestposemat = poseutil.mat((0, 0, 0, 0, 0, 0))
        self.Qbuilt = False # force rebuild in case data's changed
        if self.reciprocal:
            self.__buildindex(Q, P)
        else:
            self.__buildindex(Q)

        if self.visualise:
            self.__initvisualiser()
            self.vis.clear()
            Qvis = Q.copy()
            resample_res = 0.1
            while len(Qvis) > 1e5:
                print "resampling for visualiser..."
                import mrol_mapping.util as util
                Qvis = util.volumetricsample(Qvis, resample_res)
                resample_res = resample_res*2.
            self.vis.addmappts(Qvis)
            self.vis.setrightpts(P)
        last_dist = 9e9
        
        # get an inital association
        Qnormals = None
        Qinds, Pinds, dists, Qnormals = self.__associate(P, Q, reject_threshold, calc_normal=point2plane)
            
        Qpts, Ppts = self.get_assoc_pts(P, Q, Pinds, Qinds)
        if len(Ppts) == 0 or len(Qpts) == 0:
            print "Warning: No points within association distance"
            return np.array([0,0,0,0,0,0]), 0
            
        if self.visualise:
            import mrol_mapping.visualiser.dispxyz as dispxyz
            self.line_handles = dispxyz.plot2ptlines(Qpts,Ppts)  
            if point2plane:
                Qnormals_pts = Qpts + Qnormals
                col = np.ones(Qpts.shape)
                self.normal_handles = dispxyz.plot2ptlines(Qpts,Qnormals_pts,color=col)
        
        if not quiet:
            print "Starting at:", np.sqrt(np.mean(dists**2.))
        
        for iteration in range(iterations):
            
            if len(Ppts) == 0 or len(Qpts) == 0:
                if not quiet:
                    print "No points within association distance"
                return np.array([0,0,0,0,0,0]), 0
                
            if point2plane:
                X = self.__calctransform_normal(Ppts, Qpts, Qnormals)
            else:
                X = self.__calctransform(Ppts, Qpts)
                
            bestposemat = np.dot(bestposemat, poseutil.mat(X))
            P = poseutil.transformPoints(P, poseutil.mat(X))
            
            if self.visualise:
                self.vis.setrightpts(P)
            
            if reassoc:  # reassociate for the next iteration
                self.__buildindex(Q, P) # rebuild the index for P as it's moved
                Qinds, Pinds, dists, Qnormals = self.__associate(P, Q, reject_threshold, calc_normal=point2plane)
            
            Qpts, Ppts = self.get_assoc_pts(P, Q, Pinds, Qinds)
            dists = self.get_assoc_dists(Ppts, Qpts, Qnormals)
            dist = np.sqrt(np.mean(dists**2.))    
                
            if not quiet:
                print iteration, dist
                
            if dist < terminate_threshold: break
            
            if np.abs(dist - last_dist) < 1e-9:
                if not quiet:
                    print "Warning, no longer converging and termination threshold not reached. Possible local minima. Try changing the rejection threshold."
                break
                
            if iteration == iterations-1:
                print "Warning, maximum iterations reached."
            
            

#            Qpts, Ppts = self.get_assoc_pts(P, Q, Pinds, Qinds)
#            if len(Ppts) == 0 or len(Qpts) == 0:
#                print "Warning: No points within association distance"
#                return np.array([0,0,0,0,0,0]), 0
            
            last_dist = dist
            
            if self.visualise:
                import mrol_mapping.visualiser.dispxyz as dispxyz
                self.__vis_clear_lines(self.line_handles)
                self.line_handles = dispxyz.plot2ptlines(Qpts,Ppts)
                if point2plane:
                    self.__vis_clear_lines(self.normal_handles)
                    Qnormals_pts = Qpts + Qnormals
                    col = np.ones(Qpts.shape)
                    self.normal_handles = dispxyz.plot2ptlines(Qpts,Qnormals_pts,color=col)
                #import pdb;pdb.set_trace()
                
        bestposemat = np.dot(bestposemat, initialposemat)
        result = np.array(poseutil.posetuple(bestposemat))
        if not quiet:
            print "Final pose:", result
        #inlier_pts, outlier_pts = classify_pts(P, Pinds)
        return result, dist #, inlier_pts, outlier_pts
"""
def segment(P_aligned, Q, reject_threshold, rebuild_index=False):
    # TODO can we check if the index is already built and not rebuild if not necessary?
    if rebuild_index:
        buildindex(Q)
    Qinds, Pinds, dist = __associate(P_aligned, reject_threshold)
    inlier_pts, outlier_pts = classify_pts(P_aligned, Pinds)
    return inlier_pts, outlier_pts
    
def classify_pts(P, Pinds):
    inlier_pts = P[Pinds] 
    outlier_pts = P[np.logical_not(Pinds)]
    return inlier_pts, outlier_pts
    
def cost(pose, P, Q, threshold=0.5, rebuild=True):
    '''Cost function based on root mean squared distance to approximate nearest 
    neighbour. '''
    if rebuild:
        buildindex(Q)
    # TODO refactor this into a class so make the requirement to call 
    # buildindex less error prone
    P = poseutil.transformPoints(P, poseutil.mat(pose))
    Qinds, Pinds, dist = __associate(P, threshold)
    return dist
"""
