/*
 * Copyright 2009-2011, CSIRO.
 * Nick Hillier and Polly Alexander
 */
/*
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>. 
 */

#ifndef NOVATEL_INTERFACE_H
#define NOVATEL_INTERFACE_H

/*** Includes ***/
/* ROS stuff */
#include <ros/ros.h>
#include <ros/node_handle.h>
#include <ros/master.h>
#include <ros/time.h>
#include <ros/console.h>
/* Other ROS package stuff */
#include <sensor_msgs/NavSatFix.h>
#include <nav_msgs/Odometry.h>
#include <sensor_msgs/Imu.h>
#include <tf/transform_broadcaster.h>
#include <novatelINS/novatelINSstatus.h>
#include <novatelINS/novatelGPSstatus.h>

/* Other CSIRO stuff */
#include <small/Rotation3D.hh> // TODO: replace SMALL with ROS tf library or bullet?
//#include <Linearmath/btQuaternion.h>

/* ACFR Gearbox stuff */
#include <gbxutilacfr/exceptions.h>
#include <gbxnovatelacfr/driver.h>

/*** Definitions ***/
/* magnetic field defines for initializing azimuth from an imu3dmg */
#define MAG_STD_DEFAULT 3.0
#define MAG_AVG_SIZE 20
/* comms */
#define DEFAULT_BAUD 115200 // 200
#define DEFAULT_PORT "/dev/ttyS0"
#define DEFAULT_ALIGNMENT_METHOD "static" // Other option is "kinematic"
#define DEG2RAD(x) ((x)*M_PI/180.0)

/* namespace modifications */
namespace gna = gbxnovatelacfr;
using namespace std;

class NovatelInterfaceROS
{
	protected:
        ros::Publisher _pub_GPSFix, _pub_Odometry, _pub_INSStatus, _pub_GPSStatus;
        ros::Subscriber _sub_imu3dmg; // magnetometer is used for inital alignment calibration if alignment method is static
        tf::TransformBroadcaster _br_tf, _br_tf_gps, _br_tf_gpsoffset;
        
        boost::shared_ptr<ros::NodeHandle> _node, _node_ns;

        sensor_msgs::NavSatFix _GPSFixData;
        nav_msgs::Odometry _OdometryData;
        sensor_msgs::Imu _imu3dmgData;
        novatelINS::novatelINSstatus _novatelINSstatus;
        novatelINS::novatelGPSstatus _novatelGPSstatus;
        tf::Transform _tf_transform, _tf_gps_transform, _tf_gpsoffset_transform;
		
		std::string alignType;
		std::string port;
		std::string cntlframe_id, gpsframe_id, baseframe_id;
		int baud;
		double imuOffset[3]; // position of IMU (JAUS)
		double gpsOffset[3]; // position of GPS antenna (JAUS)
		bool align;

		std::auto_ptr<gna::Config> config;
		std::auto_ptr<gna::Driver> driver;
		
		double imuToGpsOffset[3];
		
		double magaz, magazstd; // azimuth according to magnetometer, standard deviation in the azimuth estimate from the magnetometer
		int num_mag_avg;
		double mag[2], magbias[2]; // magnetometer reading (X,Y), bias in the magnetometer reading
		
		double tfoffsetX,  tfoffsetY, tfoffsetZ; // offsets for TF

		int jaus2novatel_frame(std::vector<double> *vectorArg);
		int getInitAzimuth(std::vector<double> *vectorArg);
		void getLocalPos(double latitude, double longitude, double height, double *X, double *Y, double *Z);
		
		SMALL::Rotation3D rot3D, rot3Dgps; // internal tracking of vehicle orientation
		
	public:		
		NovatelInterfaceROS();
		~NovatelInterfaceROS();
		void imu3dmgCallback(const sensor_msgs::Imu::ConstPtr& data);
		int setupROS();
		int reset();
		int setupDriver();
		int start(); 
		int read_and_process_data();
		int setup();
			
};
		
#endif // NOVATEL_INTERFACE_H
