/*
 * Copyright 2009-2011, CSIRO.
 * Nick Hillier and Polly Alexander
 * 
 * NOTE: outputs coordinates in a JUAS-like frame convention, NOT ROS-like!
 * (i.e. North-East-Down : Forwards-Right-Down : X-Y-Z)
 */
/*
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>. 
 */

#include <novatelINS/NovatelInterface.h>
#include <novatelINS/gps_convert.h>



// TODO: replace sleep() with ros::Time equivalent

NovatelInterfaceROS::NovatelInterfaceROS() 
{
	;
}

NovatelInterfaceROS::~NovatelInterfaceROS() 
{
	;
}

void NovatelInterfaceROS::getLocalPos(double latitude, double longitude, double height, double *X, double *Y, double *Z){
	// pose is JAUS-like Northings/Eastings/Height
	double easting, northing;
	int zone;
	// TODO: The latlong2eastnorth() function needs review. Not confident it is correct.
	latlong2eastnorth(latitude, longitude, &easting, &northing, &zone);
	
	double tf_x, tf_y, tf_z;
	tf_x = northing - tfoffsetX;
	tf_y = easting - tfoffsetY;
	tf_z = height - tfoffsetZ;
	
	*(X) = tf_x;
	*(Y) = tf_y;
	*(Z) = tf_z;
}


void NovatelInterfaceROS::imu3dmgCallback(const sensor_msgs::Imu::ConstPtr& data){
	_imu3dmgData = *data;
	
	// TODO: ROS driver for the 3DM-GX? does not support magnetic field data at this time
	//mag[0] += _imu3dmgData.field[0];
	mag[0] += 1;
	//mag[1] += _imu3dmgData.field[1];
	mag[1] += 1;
	num_mag_avg++;
	
    return;
}

int NovatelInterfaceROS::setupROS() 
{    
    // TODO: publish as a transform to /tf
    
    _node.reset(new ros::NodeHandle());	
    _node_ns.reset(new ros::NodeHandle("~/config"));	
	_node_ns->param("port", port, (std::string)DEFAULT_PORT);
	_node_ns->param("baud", baud, DEFAULT_BAUD);
	_node_ns->param("imuOffsetX", imuOffset[0], 0.);
	_node_ns->param("imuOffsetY", imuOffset[1], 0.);
	_node_ns->param("imuOffsetZ", imuOffset[2], 0.);
	_node_ns->param("gpsOffsetX", gpsOffset[0], 0.);
	_node_ns->param("gpsOffsetY", gpsOffset[1], 0.);
	_node_ns->param("gpsOffsetZ", gpsOffset[2], 0.);
	_node_ns->param("alignType", alignType, (std::string)DEFAULT_ALIGNMENT_METHOD);
	_node_ns->param("magbiasX", magbias[0], 0.);
	_node_ns->param("magbiasY", magbias[1], 0.);
	_node_ns->param("mazazstd", magazstd, MAG_STD_DEFAULT);
	_node_ns->param("tfOffsetX", tfoffsetX, 0.);
	_node_ns->param("tfOffsetY", tfoffsetY, 0.);
	_node_ns->param("tfOffsetZ", tfoffsetZ, 0.);
	_node_ns->param("basel_frame_id", baseframe_id, (std::string)"Novatelworld");
	_node_ns->param("control_frame_id", cntlframe_id, (std::string)"NovatelINS");
	_node_ns->param("GPS_frame_id", gpsframe_id, (std::string)"NovatelGPS");
    _pub_GPSFix = _node->advertise<sensor_msgs::NavSatFix>("Novatel/NavSatFix", 100);
    _pub_Odometry = _node->advertise<nav_msgs::Odometry>("Novatel/odom", 100);
    _pub_INSStatus = _node->advertise<novatelINS::novatelINSstatus>("Novatel/INSstatus", 10);
    _pub_GPSStatus = _node->advertise<novatelINS::novatelGPSstatus>("Novatel/GPSstatus", 10);
    reset();
    
	return 0;
}

int NovatelInterfaceROS::reset() {
    _GPSFixData.header.seq = 0;
    _GPSFixData.header.frame_id = gpsframe_id;
    _GPSFixData.status.service = sensor_msgs::NavSatStatus::SERVICE_GPS;
    _OdometryData.header.seq = 0;
    _OdometryData.header.frame_id = cntlframe_id;
    _OdometryData.child_frame_id = baseframe_id;
    _novatelINSstatus.header.seq = 0;
    _novatelINSstatus.header.frame_id = cntlframe_id;
    _novatelGPSstatus.header.seq = 0;
    _novatelGPSstatus.header.frame_id = gpsframe_id;
    return 0;
}

int NovatelInterfaceROS::setupDriver()
{
    ROS_DEBUG("Baud: %d\n", baud);
	ROS_DEBUG("Port: %s\n", port.c_str());
    ROS_DEBUG("imu offset: %f,%f,%f\n", imuOffset[0], imuOffset[1], imuOffset[2]);
    ROS_DEBUG("gps offset: %f,%f,%f\n", gpsOffset[0], gpsOffset[1], gpsOffset[2]);
	
	// Vector from Bobcat origin to Novatel INS unit
	vector<double> insOffset(imuOffset, imuOffset + sizeof(imuOffset)/sizeof(*imuOffset));
	ROS_DEBUG("imuOffset (JAUS)= %f %f %f\n", insOffset[0], insOffset[1], insOffset[2]);
	jaus2novatel_frame(&insOffset); // Transform into Novatel frame convention
    ROS_DEBUG("imuOffset (novatel) = %f %f %f\n", insOffset[0], insOffset[1], insOffset[2]);

	// Vector from Vehicle origin to GPS antenna
	vector<double> originToGps(gpsOffset, gpsOffset + sizeof(gpsOffset)/sizeof(*gpsOffset));
	ROS_DEBUG("gpsOffset (JAUS) = %f %f %f", originToGps[0], originToGps[1], originToGps[2]);
	jaus2novatel_frame(&originToGps); // Transform into Novatel frame convention
    ROS_DEBUG("gpsOffset (novatel) = %f %f %f", originToGps[0], originToGps[1], originToGps[2]);
    
	// Vector from Novatel INS unit to GPS antenna
	vector<double> imuToGps(3,0.0); //empty vector with 3 elements
	for (int i = 0; i < 3; i++)
	{
		imuToGps[i] = originToGps[i] - insOffset[i];
	}
	ROS_DEBUG("imuToGps (novatel) = %f %f %f", imuToGps[0], imuToGps[1], imuToGps[2]);

	auto_ptr<gna::SimpleConfig > simpleCfg;
	simpleCfg.reset( new gna::SimpleConfig(port, baud, "IMU_HG1700_AG62" , imuToGps));
	config.reset( new gna::Config(*simpleCfg.get()) );

	// set up a couple of extra parameters
	config->enableInsPhaseUpdate_ = true;
	config->enableRTK_ = true;
	config->dtInsPva_ = 0.1;
	config->dtInsCov_ = 0.1; //!cov
	vector<double > fullUncertainty(3, 0.0);
	fullUncertainty[0] = 0.02;
	fullUncertainty[1] = 0.02;
	fullUncertainty[2] = 0.02;
	config->imuToGpsOffsetUncertainty_ = fullUncertainty;
	config->enableInsOffset_ = true;
	config->insOffset_ = insOffset; // read from config file
	config->enableVehicleBodyRotation_ = true;
	vector<double > bodyRot(3,0.0);
	bodyRot[0] = -2.6;
	bodyRot[1] = 0.8;
	bodyRot[2] = 0.0;
	config->vehicleBodyRotation_ = bodyRot;
	vector<double > bodyRotUncert(3, 0.0);
	bodyRotUncert[0] = 0.5; 
	bodyRotUncert[1] = 0.6;
	bodyRotUncert[2] = 0.6;
	config->vehicleBodyRotationUncertainty_ = bodyRotUncert;
    
    if ((alignType != "kinematic") && (alignType != "static")){
        ROS_WARN("Unknown alignment type, defaulting to %s\n",DEFAULT_ALIGNMENT_METHOD);
        alignType = DEFAULT_ALIGNMENT_METHOD;
    }
    if (alignType == "kinematic") {
        align = false;
    } else {// alignType == "static"
        align = true;
    }

    ROS_DEBUG("Alignment type: %s (%s)\n", alignType.c_str(), (align)?"true (static)":"false (kinematic)");

	vector<double > initAzimuth(2, 0.0);	
	if (align) {
		// static	
		// initAzimuth[0] = -90.0;
		// initAzimuth[1] = 10.0;
		ROS_INFO("Novatel static align start");
		getInitAzimuth(&initAzimuth);
		ROS_INFO("Novatel static align done");
		config->initAzimuth_ = initAzimuth;
		config->enableSetInitAzimuth_ = true;
	} else {
		// kinematic
		config->enableSetInitAzimuth_ = false;
	}

	//check that it makes sense
    int err = 0;
	if(!config->isValid()){
        err = -1;
        ROS_ERROR("NovatelInterfaceROS::setupDriver: Invalid config");
	}
    
	ROS_DEBUG("Config is: %s", config->toString().c_str());
    
	return err;
}

int NovatelInterfaceROS::start()
{
	//create the driver; this gets things rolling (opens the serial port, sets up the device)
	int errorCnt = 0;
	do{
		try{
			driver.reset( new gna::Driver( *(config.get()) ) );
		}
		catch(std::exception &e){
			ROS_WARN("%s: Waiting 5 secs before re-trying", e.what());
			//std::cout << e.what() << "\nWaiting 5 sec before re-trying.\n\n";
			sleep(5);
		}
		catch(...){
			ROS_WARN( "caught unknown exception!\nWaiting 5 sec before re-trying.\n\n");
			sleep(5);
		}
	}while(0 == driver.get() && errorCnt++<1); // try up to 5 times 
	
	//check that we succeeded
	int err = 0;
	if(0 == driver.get()){
		ROS_ERROR("Failed to set up driver!\n");
		err = -1;
	}
	return err;
}

int NovatelInterfaceROS::read_and_process_data()
{
	auto_ptr<gna::GenericData > generic;
	try{
		generic = driver->read();
	}
	catch( const std::exception& e ){
		int ins_driver_read_exception_caught = 0;
		assert(ins_driver_read_exception_caught);
		ROS_ERROR("%s: NovatelInterfaceROS::read_and_process_data: exception cought on driver read", e.what());
		return -1; 
	}
	//figure out what we've got
	switch(generic->type()){
		case gna::InsPva:
		{
			gna::InsPvaData *data = dynamic_cast<gna::InsPvaData *>(generic.get());
			if (0 == data) {
				ROS_ERROR("InsPva data structure empty");
				return -1;
			}
			
			// Set the timestamp 
			// TODO: should this be the GPS time or ROS time?
			ros::Time timestamp((int)data->timeStampSec, (int)(data->timeStampUSec*1e3)); // sec, nsec WALL TIME from driver - call to gettimeofday()
			//ros::Time::Time timestamp(,)); // sec, nsec GPS TIME
			
			// TODO: should the header completion be outside the switch, at the end with the publish?
			_OdometryData.header.stamp = timestamp; 
			_novatelINSstatus.header.stamp = timestamp; 
			
			// fill in the standard ROS messages
			double tf_x, tf_y, tf_z;
			getLocalPos((double)data->latitude, (double)data->longitude, -(double)data->height, &tf_x, &tf_y, &tf_z);
			
			_tf_transform.setOrigin(tf::Vector3(tf_x, tf_y, tf_z));
			
			_OdometryData.pose.pose.position.x = tf_x;
			_OdometryData.pose.pose.position.y = tf_y;
			_OdometryData.pose.pose.position.z = tf_z;
			
			// twist is in the vehicle frame
			//btQuaternion q;
			//q.setEulerYPR(-(double)data->azimuth,(double)data->pitch,(double)data->roll); // TODO: Is this right? fixed or relative axis?
			double roll, pitch, yaw;
			roll = (double)DEG2RAD(data->roll);
			pitch = (double)DEG2RAD(data->pitch);
			yaw = (double)DEG2RAD(-data->azimuth);
            //cout << roll << ", " << pitch << ", " << yaw << ", " << data->roll << ", " << data->pitch << ", " << -data->azimuth << endl;
			rot3D.setRollPitchYawRad(roll, pitch, yaw); // set using Rodriguez's formula as relative axis
			double quatX = rot3D.getQuaternion()[1];
			double quatY = rot3D.getQuaternion()[2];
			double quatZ = rot3D.getQuaternion()[3];
			double quatW = rot3D.getQuaternion()[0];
            
            //cout << quatX << ", " << quatY << ", " << quatZ << ", " << quatW << endl;
			
			_tf_transform.setRotation( tf::Quaternion(quatX, quatY, quatZ, quatW) );
			
			_OdometryData.pose.pose.orientation.x = quatX;
			_OdometryData.pose.pose.orientation.y = quatY;
			_OdometryData.pose.pose.orientation.z = quatZ;
			_OdometryData.pose.pose.orientation.w = quatW;
            
            
			SMALL::Vector3D velNEU, velBody;
			velNEU = (double)data->northVelocity, (double)data->eastVelocity, -(double)data->upVelocity;
			velBody = rot3D.rotate(velNEU);
			_OdometryData.twist.twist.linear.x = velBody[0];
			_OdometryData.twist.twist.linear.y = velBody[1];
			_OdometryData.twist.twist.linear.z = velBody[2];
			// Novatel unit doesn't give us angular rates
			// TODO: difference the headings to get the rates
			_OdometryData.twist.twist.angular.x = 0.0;
			_OdometryData.twist.twist.angular.y = 0.0;
			_OdometryData.twist.twist.angular.z = 0.0;
			// TODO: fill in a custom Novatel ROS msg with the remaining guff
			
			// clear the statusMessage string, overwrite all elements with 0:
			//memset(insPvaData.statusMessage, 0, MAX_STATUS_MESSAGE_LENGTH);
			
			//insPvaData.gpsWeekNr		= (int)data->gpsWeekNr	 ; 
			//insPvaData.secIntoWeek		= (double)data->secIntoWeek	 ;

			//insPvaData.statusMessageType = (int)data->statusMessageType;
			//sprintf(insPvaData.statusMessage, "%s", data->statusMessage.c_str()) ;
            
            _novatelINSstatus.statusMessage = data->statusMessage;
            _novatelINSstatus.statusType = (int)data->statusMessageType;
            _novatelINSstatus.weekNumber = (int)data->gpsWeekNr;
            _novatelINSstatus.secIntoWeek = ros::Time ((double)data->secIntoWeek);

			ROS_DEBUG("insPvaData: %s", data->toString().c_str());
			ROS_DEBUG("got inspva\n");
			
			break;
		}
		
		//!cov:start
		case gna::InsCov:
		{
			gna::InsCovData *data = dynamic_cast<gna::InsCovData *>(generic.get());
			if (0 == data) {
				ROS_ERROR("InsCov data structure empty");
				return -1;
			}
			
			// Set the timestamp 
			// TODO: should this be the GPS time or ROS time?
			//ros::Time::Time timestamp((int)data->timeStampSec, (int)(data->timeStampUSec*1e3)); // sec, nsec WALL TIME from driver - call to gettimeofday()
			//ros::Time::Time timestamp(,)); // sec, nsec GPS TIME
			
			// fill in the standard ROS message
			_OdometryData.pose.covariance[0] = (double)data->posxx;
			_OdometryData.pose.covariance[1] = (double)data->posxy;
			_OdometryData.pose.covariance[2] = (double)data->posxz;
			_OdometryData.pose.covariance[6] = (double)data->posyx;
			_OdometryData.pose.covariance[7] = (double)data->posyy;
			_OdometryData.pose.covariance[8] = (double)data->posyz;
			_OdometryData.pose.covariance[12] = (double)data->poszx;
			_OdometryData.pose.covariance[13] = (double)data->poszy;
			_OdometryData.pose.covariance[14] = (double)data->poszz;
			
			_OdometryData.pose.covariance[21] = (double)data->attxx;
			_OdometryData.pose.covariance[22] = (double)data->attxy;
			_OdometryData.pose.covariance[23] = (double)data->attxz;
			_OdometryData.pose.covariance[27] = (double)data->attyx;
			_OdometryData.pose.covariance[28] = (double)data->attyy;
			_OdometryData.pose.covariance[29] = (double)data->attyz;
			_OdometryData.pose.covariance[33] = (double)data->attzx;
			_OdometryData.pose.covariance[34] = (double)data->attzy;
			_OdometryData.pose.covariance[35] = (double)data->attzz;

			// Convert the velocity covar to the vehicle frame
			SMALL::Matrix33 velcovar, velcovarVeh;
			velcovar =  (double)data->velxx, (double)data->velxy, (double)data->velxz,
						(double)data->velyx, (double)data->velyy, (double)data->velyz,
						(double)data->velzx, (double)data->velzy, (double)data->velzz;
			//velcovarVeh = rot3D.i().getRotationMatrix()* velcovar * rot3D.i().getRotationMatrix().transposed(); // TODO : is this the same as rot3D.i().getRotationMatrix()* velcovar * rot3D.getRotationMatrix() ???
			velcovarVeh = rot3D.getRotationMatrix()* velcovar * rot3D.getRotationMatrix().transposed();
			
			_OdometryData.twist.covariance[0] = velcovarVeh(1,1);
			_OdometryData.twist.covariance[1] = velcovarVeh(1,2);
			_OdometryData.twist.covariance[2] = velcovarVeh(1,3);
			_OdometryData.twist.covariance[6] = velcovarVeh(2,1);
			_OdometryData.twist.covariance[7] = velcovarVeh(2,2);
			_OdometryData.twist.covariance[8] = velcovarVeh(2,3);
			_OdometryData.twist.covariance[12] =velcovarVeh(3,1);
			_OdometryData.twist.covariance[13] =velcovarVeh(3,2);
			_OdometryData.twist.covariance[14] =velcovarVeh(3,3);

			//insCovData.gpsWeekNr		= (int)data->gpsWeekNr; 
			//insPvaData.secIntoWeek		= (double)data->secIntoWeek;
		
			ROS_DEBUG("insCovData: %s", data->toString().c_str());
		
			ROS_DEBUG("got inscov\n");
		
			break;
		}
		//!cov:end		
		
		case gna::BestGpsPos:
		{
			gna::BestGpsPosData *data = dynamic_cast<gna::BestGpsPosData *>(generic.get());
			if (0 == data) {
				ROS_ERROR("BestGpsPos returned data == 0");
				return -1;
			}
			
			// Set the timestamp 
			// TODO: should this be the GPS time or ROS time?
			ros::Time timestamp((int)data->timeStampSec, (int)(data->timeStampUSec*1e3)); // sec, nsec WALL TIME from driver - call to gettimeofday()
			//ros::Time::Time timestamp(,)); // sec, nsec GPS TIME
			
			_GPSFixData.header.stamp = timestamp; 
            _novatelGPSstatus.header.stamp = timestamp;

			// fill in the standard ROS message
			_GPSFixData.latitude = (double)data->latitude;
			_GPSFixData.longitude = (double)data->longitude;
			_GPSFixData.altitude = (double)data->heightAMSL;
			_GPSFixData.position_covariance[0] = data->sigmaLatitude*data->sigmaLatitude;
			_GPSFixData.position_covariance[4] = data->sigmaLongitude*data->sigmaLongitude;
			_GPSFixData.position_covariance[8] = data->sigmaHeight*data->sigmaHeight;
			_GPSFixData.position_covariance_type = sensor_msgs::NavSatFix::COVARIANCE_TYPE_DIAGONAL_KNOWN;
			
			// TODO: make sense of the conversion to the NavSatStatus Solution Status
			ROS_WARN_ONCE("NavSatStatus Solution Status not properly implemented, read with caution!");
			//switch (data->positionType) {
			//	case :
			//		_GPSFixData.status.status = sensor_msgs::NavSatFix.STATUS_GBAS_FIX; // ground based augmentation
			//	default:
			//		_GPSFixData.status.status = sensor_msgs::NavSatFix.STATUS_FIX; // standard sat fix
			//}
			if (data->numL1RangesRTK > 0) {
				_GPSFixData.status.status = sensor_msgs::NavSatStatus::STATUS_GBAS_FIX;
			} else if (data->numL1Ranges > 4) {
				_GPSFixData.status.status = sensor_msgs::NavSatStatus::STATUS_FIX;
			} else {
				_GPSFixData.status.status = sensor_msgs::NavSatStatus::STATUS_NO_FIX;
			}
			
			double tf_x, tf_y, tf_z;			
			getLocalPos(_GPSFixData.latitude, _GPSFixData.longitude, -_GPSFixData.altitude, &tf_x, &tf_y, &tf_z);
			
			tf_z -= data->undulation;
			_tf_gps_transform.setOrigin(tf::Vector3(tf_x, tf_y, tf_z)); // rotation set in GPS velocity section below
			
			_tf_gpsoffset_transform.setOrigin(tf::Vector3(gpsOffset[0], gpsOffset[1], gpsOffset[2])); // Should this be "originToGps[]" or gpsOffset?
			_tf_gpsoffset_transform.setRotation( tf::Quaternion(0, 0, 0, 1) ); // Not sure what this should be...
			
			// TODO: fill in a custom Novatel ROS msg with the remaining guff			

			//gpsBestPosData.gpsWeekNr      		=data->gpsWeekNr      	;
			//gpsBestPosData.msIntoWeek     		=data->msIntoWeek     	;
			//gpsBestPosData.solutionStatus 		=data->solutionStatus 	;
			//gpsBestPosData.positionType   		=data->positionType   	;

			//gpsBestPosData.undulation     		=data->undulation     	;
			//gpsBestPosData.datumId        		=data->datumId        	;
			
			//memcpy(gpsBestPosData.baseStationId, data->baseStationId, sizeof(gpsBestPosData.baseStationId));
			//gpsBestPosData.diffAge        		=data->diffAge        	;  
			//gpsBestPosData.solutionAge    		=data->solutionAge    	; 
			//gpsBestPosData.numObservations		=data->numObservations	;
			//gpsBestPosData.numL1Ranges    		=data->numL1Ranges    	;
			//gpsBestPosData.numL1RangesRTK 		=data->numL1RangesRTK 	;
			//gpsBestPosData.numL2RangesRTK 		=data->numL2RangesRTK 	;
			//gpsBestPosData.statusMessageType	=data->statusMessageType ;
			//sprintf(gpsBestPosData.statusMessage, "%s", data->statusMessage.c_str()) ;
			//gpsBestPosData.timeStampSec     	=data->timeStampSec      ;
			//gpsBestPosData.timeStampUSec    	=data->timeStampUSec     ;
            
            _novatelGPSstatus.weekNumber = (int)data->gpsWeekNr;
            _novatelGPSstatus.secIntoWeek = ros::Time ((double)data->msIntoWeek/1e3);
            _novatelGPSstatus.GPSsolutionStatus = data->solutionStatus;
            _novatelGPSstatus.GPSpositionType = data->positionType;
            _novatelGPSstatus.GPSundulation = data->undulation;
            _novatelGPSstatus.DatumId = data->datumId;
            _novatelGPSstatus.numObservations = data->numObservations;
            _novatelGPSstatus.numL1Ranges = data->numL1Ranges;
            _novatelGPSstatus.numL1RangesRTK = data->numL1RangesRTK;
            _novatelGPSstatus.numL2RangesRTK = data->numL2RangesRTK;

			ROS_DEBUG("bestGpsPosData: %s", data->toString().c_str());
		
			ROS_DEBUG("got bestGpsPos\n");
			
			break;
		}
		case gna::BestGpsVel:
		{
			gna::BestGpsVelData *data = dynamic_cast<gna::BestGpsVelData *>(generic.get());
			if (0 == data) {
				ROS_ERROR("BestGpsVel returned data == 0");
				return -1;
			}
			ROS_WARN_ONCE("Handling for GPS velocity data not yet implemented");
			// Set the timestamp 
			// TODO: should this be the GPS time or ROS time?
			ros::Time timestamp((int)data->timeStampSec, (int)(data->timeStampUSec*1e3)); // sec, nsec WALL TIME from driver - call to gettimeofday()
			//ros::Time::Time timestamp(,)); // sec, nsec GPS TIME
			double GPS_yaw = DEG2RAD(-data->trackOverGround);
			//cout << GPS_yaw  << endl;
			rot3Dgps.setRollPitchYawRad(0., 0., GPS_yaw); // no roll or pitch info from the GPS
			double quatX = rot3Dgps.getQuaternion()[1];
			double quatY = rot3Dgps.getQuaternion()[2];
			double quatZ = rot3Dgps.getQuaternion()[3];
			double quatW = rot3Dgps.getQuaternion()[0];
			
			_tf_gps_transform.setRotation( tf::Quaternion(quatX, quatY, quatZ, quatW) );

			//gpsBestVelData.gpsWeekNr      	= data->gpsWeekNr; 
			//gpsBestVelData.msIntoWeek     	= data->msIntoWeek;
			//gpsBestVelData.solutionStatus 	= data->solutionStatus;
			//gpsBestVelData.velocityType   	= data->velocityType; 
			//gpsBestVelData.latency        	= data->latency;               
			//gpsBestVelData.diffAge        	= data->diffAge;              
			//gpsBestVelData.horizontalSpeed 	= data->horizontalSpeed;     
			//gpsBestVelData.trackOverGround 	= data->trackOverGround;    
			//gpsBestVelData.verticalSpeed  	= data->verticalSpeed;     
			//gpsBestVelData.statusMessageType = data->statusMessageType;     
			//sprintf(gpsBestVelData.statusMessage, "%s", data->statusMessage.c_str()) ;
			//gpsBestVelData.timeStampSec   	= data->timeStampSec; 
			//gpsBestVelData.timeStampUSec  	= data->timeStampUSec;

			ROS_DEBUG("bestGpsVelData: %s", data->toString().c_str());
		
			ROS_DEBUG("got bestGpsVel\n");
			
			break;
		}
		case gna::RawImu:
		{
			//gna::RawImuData *data = dynamic_cast<gna::RawImuData *>(generic.get());

			//if (0 == data) {
			//	ROS_ERROR("RawImu returned data == 0");
			//	return -1;
			//}
			//
			//// Set the timestamp 
			//// TODO: should this be the GPS time or ROS time?
			//ros::Time::Time timestamp((int)data->timeStampSec, (int)(data->timeStampUSec*1e3)); // sec, nsec WALL TIME from driver - call to gettimeofday()
			////ros::Time::Time timestamp(,)); // sec, nsec GPS TIME
			//
			//rawIMUData.gpsWeekNr       = data->gpsWeekNr; 
			//rawIMUData.secIntoWeek     = data->secIntoWeek;
			//rawIMUData.zDeltaV         = data->zDeltaV;  
			//rawIMUData.yDeltaV         = data->yDeltaV;
			//rawIMUData.xDeltaV         = data->xDeltaV; 
			//rawIMUData.zDeltaAng       = data->zDeltaAng;
			//rawIMUData.yDeltaAng       = data->yDeltaAng;
			//rawIMUData.xDeltaAng       = data->xDeltaAng;
			//rawIMUData.statusMessageType = data->statusMessageType;
			//sprintf(rawIMUData.statusMessage, "%s", data->statusMessage.c_str()) ;
			//rawIMUData.timeStampSec    = data->timeStampSec;
			//rawIMUData.timeStampUSec   = data->timeStampUSec; 
			//
			//if (!rawIMU.t_writefrom(rawIMUData)) {
			//	ROS_WARN("rawIMU.t_writefrom returned false");
			//}
            //
			//
			//ROS_DEBUG("rawIMU Data: %s", data->toString().c_str());
			ROS_WARN_ONCE("Handling for raw IMU data not yet implemented");
			//ROS_DEBUG("got rawIMU \n");
			
			break;
		}
		default:
		{
			ROS_WARN("Got unknown message\n");
			return -1;
			break;
		}
	}
	_pub_GPSFix.publish(_GPSFixData);
	_pub_Odometry.publish(_OdometryData);
    _pub_INSStatus.publish(_novatelINSstatus);
    _pub_GPSStatus.publish(_novatelGPSstatus);
	_br_tf.sendTransform(tf::StampedTransform(_tf_transform, _OdometryData.header.stamp, baseframe_id, cntlframe_id)); // INS control point
	_br_tf_gps.sendTransform(tf::StampedTransform(_tf_gps_transform, _GPSFixData.header.stamp, baseframe_id, gpsframe_id)); // GPS antennae
	_br_tf_gpsoffset.sendTransform(tf::StampedTransform(_tf_gpsoffset_transform, _GPSFixData.header.stamp, gpsframe_id, (cntlframe_id + (std::string)"_gps"))); // GPS control point
	
	return 0;
}

/* This funtion transforms a vector from a JAUS compliant coordinate frame to
 * the Novatel convention used by the INS unit. */
int NovatelInterfaceROS::jaus2novatel_frame(vector<double> *vectorArg)
{
	double temp = vectorArg->at(0);
	vectorArg->at(0) = vectorArg->at(1); // x(Novatel) = y(Jaus)
	vectorArg->at(1) = temp; // y(Novatel) = x(Jaus)
	vectorArg->at(2) = - vectorArg->at(2); // z(Novatel) = - z(Jaus)
	return 0;
}

int NovatelInterfaceROS::getInitAzimuth(vector<double> *vectorArg)
{
	ROS_WARN("Magnetic field static alignment is not currently implemented.");
	// TODO: allow static alignment in any direction by using a magnetometer
	// Code below and in the imu3dmgCallback() is how to do it.
	
	mag[0] = 0.0; mag[1] = 0.0;
	num_mag_avg = 0;
	_sub_imu3dmg = _node->subscribe("imu3DMG", 1, &NovatelInterfaceROS::imu3dmgCallback, this);
	
	
	ROS_WARN("Skipping magnetic calibration!");
	num_mag_avg = MAG_AVG_SIZE; // TODO: remove this line when using imu for magnetic heading determination
	
	
	
	
	ROS_DEBUG("Microstrain ave start");	
	while (num_mag_avg < MAG_AVG_SIZE) { // the call back increments num_mag_avg
		// do nothing
		usleep(100000);
	}
	num_mag_avg--;
	mag[0] /= num_mag_avg; mag[1] /= num_mag_avg; 
	mag[0] -= magbias[0]; mag[1] -= magbias[1];
	// TODO: calculate the standard deviation somewhere too.
	ROS_DEBUG("Microstrain ave done");
	
	magaz = atan2(mag[1], mag[0])*180/M_PI;
	if (magaz < 0) {
		magaz += 360.0;
	};
	vectorArg->at(0) = magaz;
	vectorArg->at(1) = magazstd;
	
	ROS_WARN("Setting static alignment to North, if the novatel is not facing North, then you need to re-orient and re-start the driver.");
	vectorArg->at(0) = 0.0; // North
	vectorArg->at(1) = magazstd;
	
	return num_mag_avg; // number of data points averaged over
}

int NovatelInterfaceROS::setup() {
	if(setupROS() != 0) {
		ROS_ERROR("setupROS: Failed");
		return -1;
	}	
	ROS_INFO("setupROS done");
	if(setupDriver() != 0) {
		ROS_ERROR("setupDriver: Failed");
		return -1;
	}
	ROS_INFO("setupDriver done");
	return 0;
}

int main(int argc, char * argv[])
{
	ros::init(argc, argv, "novatel");
	sleep(2.);
	NovatelInterfaceROS novInt;	
	
	if (novInt.setup() != 0) {
		ROS_FATAL("Driver setup: Failed");
		exit(-1);
	}
	ROS_INFO("Driver setup done");
	if (novInt.start() != 0) {
		ROS_FATAL("startDriver: Failed");
		exit(-1);
	}
	ROS_INFO("Driver started");
	while(ros::ok()) {
		if (novInt.read_and_process_data() != 0) {
			ROS_ERROR("read_and_process_data: Failed");
		}
		ros::spinOnce();
	}
	ROS_INFO("Exiting...");
	return 0;
}
